# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

"cli tests"

import base64
import contextlib
import io
import os
import shutil
import subprocess
import sys
import tempfile
import unittest
from unittest.mock import patch

from kapitan.cli import main
from kapitan.refs.secrets import vaultkv

REFS_PATH = tempfile.mkdtemp()

# set GNUPGHOME if only running this test
# otherwise it will reuse the value from test_gpg.py
if os.environ.get("GNUPGHOME", None) is None:
    GNUPGHOME = tempfile.mkdtemp()
    os.environ["GNUPGHOME"] = GNUPGHOME


class CliFuncsTest(unittest.TestCase):
    def setUp(self):
        example_key = "examples/kubernetes/refs/example@kapitan.dev.key"
        example_key = os.path.join(os.getcwd(), example_key)
        example_key_ownertrust = tempfile.mktemp()

        # always trust this key - for testing only!
        with open(example_key_ownertrust, "w") as fp:
            fp.write("D9234C61F58BEB3ED8552A57E28DC07A3CBFAE7C:6\n")

        subprocess.run(["gpg", "--import", example_key])
        subprocess.run(["gpg", "--import-ownertrust", example_key_ownertrust])
        os.remove(example_key_ownertrust)

    def test_cli_secret_write_reveal_gpg(self):
        """
        run $ kapitan refs --write gpg:test_secret
        and $ kapitan refs --reveal
        with example@kapitan.dev recipient
        """
        test_secret_content = "I am a secret!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "gpg:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
            "--recipients",
            "example@kapitan.dev",
        ]
        main()

        test_tag_content = "revealing: ?{gpg:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content), stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_base64_write_reveal_gpg(self):
        """
        run $ kapitan refs --write gpg:test_secretb64 --base64
        and $ kapitan refs --reveal
        with example@kapitan.dev recipient
        """
        test_secret_content = "I am another secret!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "gpg:test_secretb64",
            "-f",
            test_secret_file,
            "--base64",
            "--refs-path",
            REFS_PATH,
            "--recipients",
            "example@kapitan.dev",
        ]
        main()

        test_tag_content = "?{gpg:test_secretb64}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        stdout_base64 = base64.b64decode(stdout.getvalue()).decode()
        self.assertEqual(test_secret_content, stdout_base64)

        os.remove(test_tag_file)

    def test_cli_ref_reveal_recursive_dir(self):
        """
        run $ kapitan refs --reveal -f /some/dir
        where /some/dir has manifests in nested dirs:
        /some/dir/1.yml
        /some/dir/another/2.yml
        /some/dir/another/dir/3.yml
        """
        # create 3 refs
        for ref_count in range(1, 4):
            ref_content = "I am ref{}!".format(ref_count)
            ref_file = tempfile.mktemp()
            with open(ref_file, "w") as fp:
                fp.write(ref_content)

            sys.argv = [
                "kapitan",
                "refs",
                "--write",
                "base64:test_ref_{}".format(ref_count),
                "-f",
                ref_file,
                "--refs-path",
                REFS_PATH,
            ]
            main()

        # create nested dir structure with unrevealed manifests
        unrevealed_dir = tempfile.mkdtemp()
        ref_content = """---\nref_value_{}: {}\n"""

        some_dir = os.path.join(unrevealed_dir, "some/dir")
        some_dir_other = os.path.join(unrevealed_dir, "some/dir/another")
        some_dir_another = os.path.join(unrevealed_dir, "some/dir/another/dir")
        os.makedirs(some_dir)
        os.makedirs(some_dir_other)
        os.makedirs(some_dir_another)

        # write manifests in nested dirs
        expected_output = ""
        for dir_path in enumerate((some_dir, some_dir_other, some_dir_another), 1):
            count, path = dir_path
            manifest_path = os.path.join(path, "{}.yml".format(count))
            with open(manifest_path, "w") as f:
                ref = "?{{base64:test_ref_{}}}".format(count)
                f.write(ref_content.format(count, ref))

            # set expected revealed output
            expected_ref_rev = "I am ref{}!".format(count)
            expected_output += ref_content.format(count, expected_ref_rev)

        sys.argv = ["kapitan", "refs", "--reveal", "-f", some_dir, "--refs-path", REFS_PATH]
        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual(expected_output, stdout.getvalue())

    def test_cli_secret_validate_targets(self):
        """
        run $ kapitan refs --validate-targets
        expect 0 (success) exit status code
        """
        with self.assertRaises(SystemExit) as cm:
            sys.argv = [
                "kapitan",
                "refs",
                "--validate-targets",
                "--refs-path",
                "examples/kubernetes/refs/targets/",
                "--inventory-path",
                "examples/kubernetes/inventory/",
            ]
            main()
        self.assertEqual(cm.exception.code, 0)

    def test_cli_secret_write_reveal_gkms(self):
        """
        run $ kapitan refs --write gkms:test_secret
        and $ kapitan refs --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "gkms:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
            "--key",
            "mock",
        ]
        main()

        test_tag_content = "revealing: ?{gkms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content), stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_reveal_awskms(self):
        """
        run $ kapitan refs --write awskms:test_secret
        and $ kapitan refs --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "awskms:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
            "--key",
            "mock",
        ]
        main()

        test_tag_content = "revealing: ?{awskms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content), stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_plain_ref(self):
        """
        run $ kapitan refs --write plain:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "plain:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
        ]
        main()

        test_tag_content = "revealing: ?{plain:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content), stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_base64_ref(self):
        """
        run $ kapitan refs --write base64:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "base64:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
        ]
        main()

        test_tag_content = "revealing: ?{base64:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content), stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_base64_encoded_ref(self):
        """
        run $ kapitan refs --write base64:test_secret --base64
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_content_b64 = base64.b64encode(test_secret_content.encode())
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "base64:test_secret",
            "--base64",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
        ]
        main()

        test_tag_content = "revealing: ?{base64:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content_b64.decode()), stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_subvar_base64_ref(self):
        """
        run $ kapitan refs --write base64:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = """
        var1:
          var2: hello
        var3:
          var4: world
        """
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "base64:test_secret_subvar",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
        ]
        main()

        test_tag_content = """
        revealing1: ?{base64:test_secret_subvar@var1.var2}
        revealing2: ?{base64:test_secret_subvar@var3.var4}
        """
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()

        expected = """
        revealing1: {}
        revealing2: {}
        """
        self.assertEqual(expected.format("hello", "world"), stdout.getvalue())
        os.remove(test_tag_file)

    def test_cli_secret_subvar_gpg(self):
        """
        run $ kapitan refs --write gpg:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = """
                var1:
                  var2: hello
                var3:
                  var4: world
                """
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "gpg:test_secret_subvar",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
            "--recipients",
            "example@kapitan.dev",
        ]
        main()

        test_tag_content = """
                revealing1: ?{gpg:test_secret_subvar@var1.var2}
                revealing2: ?{gpg:test_secret_subvar@var3.var4}
                """
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()

        expected = """
                revealing1: {}
                revealing2: {}
                """
        self.assertEqual(expected.format("hello", "world"), stdout.getvalue())
        os.remove(test_tag_file)

    @patch.object(vaultkv.VaultSecret, "_decrypt")
    def test_cli_secret_write_vault(self, mock_reveal):
        """
        run $ kapitan refs --write vaultkv:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "foo:secret_test_key"
        test_secret_content_value = "secret_value"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = [
            "kapitan",
            "refs",
            "--write",
            "vaultkv:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
            "--vault-auth",
            "token",
        ]
        main()

        test_tag_content = "revealing: ?{vaultkv:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)

        mock_reveal.return_value = test_secret_content_value
        sys.argv = ["kapitan", "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {value}".format(value=test_secret_content_value), stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_searchvar(self):
        """
        run $ kapitan searchvar mysql.replicas
        """
        sys.argv = [
            "kapitan",
            "searchvar",
            "mysql.replicas",
            "--inventory-path",
            "examples/kubernetes/inventory/",
        ]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("examples/kubernetes/inventory/targets/minikube-mysql.yml   1\n", stdout.getvalue())

    def test_cli_inventory(self):
        """
        run $ kapitan inventory -t minikube-es -F -p cluster
        """
        sys.argv = [
            "kapitan",
            "inventory",
            "-t",
            "minikube-es",
            "-F",
            "-p",
            "cluster",
            "--inventory-path",
            "examples/kubernetes/inventory/",
        ]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("id: minikube\nname: minikube\ntype: minikube\nuser: minikube\n", stdout.getvalue())

    def tearDown(self):
        shutil.rmtree(REFS_PATH, ignore_errors=True)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os, sys
from .version import RECLASS_NAME

# defaults for the command-line options
OPT_STORAGE_TYPE = 'yaml_fs'
OPT_INVENTORY_BASE_URI = os.path.join('/etc', RECLASS_NAME)
OPT_NODES_URI = 'nodes'
OPT_CLASSES_URI = 'classes'
OPT_PRETTY_PRINT = True
OPT_GROUP_ERRORS = True
OPT_COMPOSE_NODE_NAME = False
OPT_NO_REFS = False
OPT_OUTPUT = 'yaml'

OPT_IGNORE_CLASS_NOTFOUND = False
OPT_IGNORE_CLASS_NOTFOUND_REGEXP = ['.*']
OPT_IGNORE_CLASS_NOTFOUND_WARNING = True

OPT_IGNORE_OVERWRITTEN_MISSING_REFERENCES = True
OPT_STRICT_CONSTANT_PARAMETERS = True

OPT_ALLOW_SCALAR_OVER_DICT = False
OPT_ALLOW_SCALAR_OVER_LIST = False
OPT_ALLOW_LIST_OVER_SCALAR = False
OPT_ALLOW_DICT_OVER_SCALAR = False
OPT_ALLOW_NONE_OVERRIDE = False

OPT_INVENTORY_IGNORE_FAILED_NODE = False
OPT_INVENTORY_IGNORE_FAILED_RENDER = False

CONFIG_FILE_SEARCH_PATH = [os.getcwd(),
                           os.path.expanduser('~'),
                           OPT_INVENTORY_BASE_URI,
                           os.path.dirname(sys.argv[0])
                          ]
CONFIG_FILE_NAME = RECLASS_NAME + '-config.yml'

REFERENCE_SENTINELS = ('${', '}')
EXPORT_SENTINELS = ('$[', ']')
PARAMETER_INTERPOLATION_DELIMITER = ':'
PARAMETER_DICT_KEY_OVERRIDE_PREFIX = '~'
PARAMETER_DICT_KEY_CONSTANT_PREFIX = '='
ESCAPE_CHARACTER = '\\'

AUTOMATIC_RECLASS_PARAMETERS = True
DEFAULT_ENVIRONMENT = 'base'



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

import base64
import datetime
import glob
import logging
import os
import re
import time
import types
from importlib import util
from random import Random, shuffle

import yaml
from kapitan import cached, defaults, utils
from kapitan.errors import CompileError
from six import string_types

logger = logging.getLogger(__name__)


def load_jinja2_filters(env):
    """Load Jinja2 custom filters into env"""
    env.filters["sha256"] = utils.sha256_string
    env.filters["b64encode"] = base64_encode
    env.filters["b64decode"] = base64_decode
    env.filters["yaml"] = to_yaml
    env.filters["fileglob"] = fileglob
    env.filters["bool"] = to_bool
    env.filters["to_datetime"] = to_datetime
    env.filters["strftime"] = strftime
    env.filters["regex_replace"] = regex_replace
    env.filters["regex_escape"] = regex_escape
    env.filters["regex_search"] = regex_search
    env.filters["regex_findall"] = regex_findall
    env.filters["reveal_maybe"] = reveal_maybe
    env.filters["ternary"] = ternary
    env.filters["shuffle"] = randomize_list


def load_module_from_path(env, path):
    """
    Loads a python module from provided path and adds it to jinja2 environment
    filter name is same as that of function
    """
    try:
        module_name = os.path.basename(path).split(".")[0]
        custom_filter_spec = util.spec_from_file_location(module_name, path)
        custom_filter_module = util.module_from_spec(custom_filter_spec)
        custom_filter_spec.loader.exec_module(custom_filter_module)
        for function in dir(custom_filter_module):
            if isinstance(getattr(custom_filter_module, function), types.FunctionType):
                logger.debug("custom filter loaded from {}".format(path))
                env.filters[function] = getattr(custom_filter_module, function)
    except Exception as e:
        logger.debug("failed to find custom filter from path {}".format(path))
        raise IOError("jinja2 failed to render, could not load filter at {}: {}".format(path, e))


def load_jinja2_filters_from_file(env, jinja2_filters):
    """
    if filter points to default file and in case it doesn't exist then proceed silently, no error
    else try to load module (which will throw error in case of non existence of file)
    """
    jinja2_filters = os.path.normpath(jinja2_filters)
    if jinja2_filters == defaults.DEFAULT_JINJA2_FILTERS_PATH:
        if not os.path.isfile(jinja2_filters):
            return
    load_module_from_path(env, jinja2_filters)


# Custom filters
def reveal_maybe(ref_tag):
    "Will reveal ref_tag if valid and --reveal flag is used"
    if cached.args["compile"].reveal:
        return cached.revealer_obj.reveal_raw(ref_tag)
    else:
        return ref_tag


def base64_encode(string):
    return base64.b64encode(string.encode("UTF-8")).decode("UTF-8")


def base64_decode(string):
    return base64.b64decode(string).decode("UTF-8")


def to_yaml(obj):
    return yaml.safe_dump(obj, default_flow_style=False)


# Following filters are from https://github.com/ansible/ansible/blob/devel/lib/ansible/plugins/filter/core.py
def fileglob(pathname):
    """return list of matched regular files for glob"""
    return [g for g in glob.glob(pathname) if os.path.isfile(g)]


def to_bool(a):
    """return a bool for the arg"""
    if a is None or isinstance(a, bool):
        return a
    if isinstance(a, string_types):
        a = a.lower()
    if a in ("yes", "on", "1", "true", 1):
        return True
    return False


def to_datetime(string, format="%Y-%m-%d %H:%M:%S"):
    return datetime.datetime.strptime(string, format)


def strftime(string_format, second=None):
    """return current date string for format. See https://docs.python.org/3/library/time.html#time.strftime for format"""
    if second is not None:
        try:
            second = int(second)
        except Exception:
            raise CompileError("Invalid value for epoch value ({})".format(second))
    return time.strftime(string_format, time.localtime(second))


def regex_replace(value="", pattern="", replacement="", ignorecase=False):
    """Perform a `re.sub` returning a string"""
    if ignorecase:
        flags = re.I
    else:
        flags = 0
    _re = re.compile(pattern, flags=flags)
    return _re.sub(replacement, value)


def regex_escape(string):
    """Escape all regular expressions special characters from STRING."""
    return re.escape(string)


def regex_search(value, regex, *args, **kwargs):
    """Perform re.search and return the list of matches or a backref"""
    groups = list()
    for arg in args:
        if arg.startswith("\\g"):
            match = re.match(r"\\g<(\S+)>", arg).group(1)
            groups.append(match)
        elif arg.startswith("\\"):
            match = int(re.match(r"\\(\d+)", arg).group(1))
            groups.append(match)
        else:
            raise CompileError("Unknown argument")

    flags = 0
    if kwargs.get("ignorecase"):
        flags |= re.I
    if kwargs.get("multiline"):
        flags |= re.M

    match = re.search(regex, value, flags)
    if match:
        if not groups:
            return match.group()
        else:
            items = list()
            for item in groups:
                items.append(match.group(item))
            return items


def regex_findall(value, regex, multiline=False, ignorecase=False):
    """Perform re.findall and return the list of matches"""
    flags = 0
    if ignorecase:
        flags |= re.I
    if multiline:
        flags |= re.M
    return re.findall(regex, value, flags)


def ternary(value, true_val, false_val, none_val=None):
    """value ? true_val : false_val"""
    if value is None and none_val is not None:
        return none_val
    elif bool(value):
        return true_val
    else:
        return false_val


def randomize_list(mylist, seed=None):
    try:
        mylist = list(mylist)
        if seed:
            r = Random(seed)
            r.shuffle(mylist)
        else:
            shuffle(mylist)
    except Exception:
        pass
    return mylist



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import yaml, os, optparse, posix, sys

from . import errors, get_path_mangler
from .defaults import *
from .constants import MODE_NODEINFO, MODE_INVENTORY


def make_db_options_group(parser, defaults={}):
    ret = optparse.OptionGroup(parser, 'Database options',
                               'Configure from where {0} collects data'.format(parser.prog))
    ret.add_option('-s', '--storage-type', dest='storage_type',
                   default=defaults.get('storage_type', OPT_STORAGE_TYPE),
                   help='the type of storage backend to use [%default]')
    ret.add_option('-b', '--inventory-base-uri', dest='inventory_base_uri',
                   default=defaults.get('inventory_base_uri', OPT_INVENTORY_BASE_URI),
                   help='the base URI to prepend to nodes and classes [%default]'),
    ret.add_option('-u', '--nodes-uri', dest='nodes_uri',
                   default=defaults.get('nodes_uri', OPT_NODES_URI),
                   help='the URI to the nodes storage [%default]'),
    ret.add_option('-c', '--classes-uri', dest='classes_uri',
                   default=defaults.get('classes_uri', OPT_CLASSES_URI),
                   help='the URI to the classes storage [%default]'),
    ret.add_option('-z', '--ignore-class-notfound', dest='ignore_class_notfound',
                   default=defaults.get('ignore_class_notfound', OPT_IGNORE_CLASS_NOTFOUND),
                   help='decision for not found classes [%default]')
    ret.add_option('-a', '--compose-node-name', dest='compose_node_name', action="store_true",
                   default=defaults.get('compose_node_name', OPT_COMPOSE_NODE_NAME),
                   help='Add subdir when generating node names. [%default]')
    ret.add_option('-x', '--ignore-class-notfound-regexp', dest='ignore_class_notfound_regexp',
                   default=defaults.get('ignore_class_notfound_regexp', OPT_IGNORE_CLASS_NOTFOUND_REGEXP),
                   help='regexp for not found classes [%default]')
    return ret


def make_output_options_group(parser, defaults={}):
    ret = optparse.OptionGroup(parser, 'Output options',
                               'Configure the way {0} prints data'.format(parser.prog))
    ret.add_option('-o', '--output', dest='output',
                   default=defaults.get('output', OPT_OUTPUT),
                   help='output format (yaml or json) [%default]')
    ret.add_option('-y', '--pretty-print', dest='pretty_print', action="store_true",
                   default=defaults.get('pretty_print', OPT_PRETTY_PRINT),
                   help='try to make the output prettier [%default]')
    ret.add_option('-r', '--no-refs', dest='no_refs', action="store_true",
                   default=defaults.get('no_refs', OPT_NO_REFS),
                   help='output all key values do not use yaml references [%default]')
    ret.add_option('-1', '--single-error', dest='group_errors', action="store_false",
                   default=defaults.get('group_errors', OPT_GROUP_ERRORS),
                   help='throw errors immediately instead of grouping them together')
    ret.add_option('-0', '--multiple-errors', dest='group_errors', action="store_true",
                   help='were possible report any errors encountered as a group')
    return ret


def make_modes_options_group(parser, inventory_shortopt, inventory_longopt,
                             inventory_help, nodeinfo_shortopt,
                             nodeinfo_longopt, nodeinfo_dest, nodeinfo_help):

    def _mode_checker_cb(option, opt_str, value, parser):
        if hasattr(parser.values, 'mode'):
            raise optparse.OptionValueError('Cannot specify multiple modes')

        if option == parser.get_option(nodeinfo_longopt):
            setattr(parser.values, 'mode', MODE_NODEINFO)
            setattr(parser.values, nodeinfo_dest, value)
        else:
            setattr(parser.values, 'mode', MODE_INVENTORY)
            setattr(parser.values, nodeinfo_dest, None)

    ret = optparse.OptionGroup(parser, 'Modes',
                               'Specify one of these to determine what to do.')
    ret.add_option(inventory_shortopt, inventory_longopt,
                   action='callback', callback=_mode_checker_cb,
                   help=inventory_help)
    ret.add_option(nodeinfo_shortopt, nodeinfo_longopt,
                   default=None, dest=nodeinfo_dest, type='string',
                   action='callback', callback=_mode_checker_cb,
                   help=nodeinfo_help)
    return ret


def make_parser_and_checker(name, version, description,
                            inventory_shortopt='-i',
                            inventory_longopt='--inventory',
                            inventory_help='output the entire inventory',
                            nodeinfo_shortopt='-n',
                            nodeinfo_longopt='--nodeinfo',
                            nodeinfo_dest='nodename',
                            nodeinfo_help='output information for a specific node',
                            add_options_cb=None,
                            defaults={}):

    parser = optparse.OptionParser(version=version)
    parser.prog = name
    parser.version = version
    parser.description = description.capitalize()
    parser.usage = '%prog [options] ( {0} | {1} {2} )'.format(inventory_longopt,
                                                             nodeinfo_longopt,
                                                             nodeinfo_dest.upper())
    parser.epilog = 'Exactly one mode has to be specified.'

    db_group = make_db_options_group(parser, defaults)
    parser.add_option_group(db_group)

    output_group = make_output_options_group(parser, defaults)
    parser.add_option_group(output_group)

    if callable(add_options_cb):
        add_options_cb(parser, defaults)

    modes_group = make_modes_options_group(parser, inventory_shortopt,
                                           inventory_longopt, inventory_help,
                                           nodeinfo_shortopt,
                                           nodeinfo_longopt, nodeinfo_dest,
                                           nodeinfo_help)
    parser.add_option_group(modes_group)

    def option_checker(options, args):
        if len(args) > 0:
            parser.error('No arguments allowed')
        elif not hasattr(options, 'mode') \
                or options.mode not in (MODE_NODEINFO, MODE_INVENTORY):
            parser.error('You need to specify exactly one mode '\
                         '({0} or {1})'.format(inventory_longopt,
                                               nodeinfo_longopt))
        elif options.mode == MODE_NODEINFO \
                and not getattr(options, nodeinfo_dest, None):
            parser.error('Mode {0} needs {1}'.format(nodeinfo_longopt,
                                                     nodeinfo_dest.upper()))
        elif options.inventory_base_uri is None and options.nodes_uri is None:
            parser.error('Must specify --inventory-base-uri or --nodes-uri')
        elif options.inventory_base_uri is None and options.classes_uri is None:
            parser.error('Must specify --inventory-base-uri or --classes-uri')

    return parser, option_checker


def get_options(name, version, description,
                            inventory_shortopt='-i',
                            inventory_longopt='--inventory',
                            inventory_help='output the entire inventory',
                            nodeinfo_shortopt='-n',
                            nodeinfo_longopt='--nodeinfo',
                            nodeinfo_dest='nodename',
                            nodeinfo_help='output information for a specific node',
                            add_options_cb=None,
                            defaults={}):

    parser, checker = make_parser_and_checker(name, version, description,
                                              inventory_shortopt,
                                              inventory_longopt,
                                              inventory_help,
                                              nodeinfo_shortopt,
                                              nodeinfo_longopt, nodeinfo_dest,
                                              nodeinfo_help,
                                              add_options_cb,
                                              defaults=defaults)
    options, args = parser.parse_args()
    checker(options, args)

    path_mangler = get_path_mangler(options.storage_type)
    options.nodes_uri, options.classes_uri = path_mangler(options.inventory_base_uri, options.nodes_uri, options.classes_uri)

    return options


def vvv(msg):
    #print(msg, file=sys.stderr)
    pass


def find_and_read_configfile(filename=CONFIG_FILE_NAME,
                             dirs=CONFIG_FILE_SEARCH_PATH):
    for d in dirs:
        f = os.path.join(d, filename)
        if os.access(f, os.R_OK):
            vvv('Using config file: {0}'.format(str(f)))
            return yaml.safe_load(open(f))
        elif os.path.isfile(f):
            raise PermissionsError('cannot read %s' % f)
    return {}



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.datatypes import Classes
from reclass.datatypes.classes import INVALID_CHARACTERS_FOR_CLASSNAMES
import unittest

try:
    import unittest.mock as mock
except ImportError:
    import mock
from reclass.errors import InvalidClassnameError

TESTLIST1 = ['one', 'two', 'three']
TESTLIST2 = ['red', 'green', 'blue']

#TODO: mock out the underlying list

class TestClasses(unittest.TestCase):

    def test_len_empty(self):
        with mock.patch.object(Classes, 'merge_unique') as m:
            c = Classes()
            self.assertEqual(len(c), 0)
            self.assertFalse(m.called)

    def test_constructor(self):
        with mock.patch.object(Classes, 'merge_unique') as m:
            c = Classes(TESTLIST1)
            m.assert_called_once_with(TESTLIST1)

    def test_equality_list_empty(self):
        self.assertEqual(Classes(), [])

    def test_equality_list(self):
        self.assertEqual(Classes(TESTLIST1), TESTLIST1)

    def test_equality_instance_empty(self):
        self.assertEqual(Classes(), Classes())

    def test_equality_instance(self):
        self.assertEqual(Classes(TESTLIST1), Classes(TESTLIST1))

    def test_inequality(self):
        self.assertNotEqual(Classes(TESTLIST1), Classes(TESTLIST2))

    def test_construct_duplicates(self):
        c = Classes(TESTLIST1 + TESTLIST1)
        self.assertSequenceEqual(c, TESTLIST1)

    def test_append_if_new(self):
        c = Classes()
        c.append_if_new(TESTLIST1[0])
        self.assertEqual(len(c), 1)
        self.assertSequenceEqual(c, TESTLIST1[:1])

    def test_append_if_new_duplicate(self):
        c = Classes(TESTLIST1)
        c.append_if_new(TESTLIST1[0])
        self.assertEqual(len(c), len(TESTLIST1))
        self.assertSequenceEqual(c, TESTLIST1)

    def test_append_if_new_nonstring(self):
        c = Classes()
        with self.assertRaises(TypeError):
            c.append_if_new(0)

    def test_append_invalid_characters(self):
        c = Classes()
        invalid_name = ' '.join(('foo', 'bar'))
        with self.assertRaises(InvalidClassnameError) as e:
            c.append_if_new(invalid_name)
        self.assertEqual(e.exception.message, "Invalid character ' ' in class name 'foo bar'.")

    def test_merge_unique(self):
        c = Classes(TESTLIST1)
        c.merge_unique(TESTLIST2)
        self.assertSequenceEqual(c, TESTLIST1 + TESTLIST2)

    def test_merge_unique_duplicate1_list(self):
        c = Classes(TESTLIST1)
        c.merge_unique(TESTLIST1)
        self.assertSequenceEqual(c, TESTLIST1)

    def test_merge_unique_duplicate1_instance(self):
        c = Classes(TESTLIST1)
        c.merge_unique(Classes(TESTLIST1))
        self.assertSequenceEqual(c, TESTLIST1)

    def test_merge_unique_duplicate2_list(self):
        c = Classes(TESTLIST1)
        c.merge_unique(TESTLIST2 + TESTLIST2)
        self.assertSequenceEqual(c, TESTLIST1 + TESTLIST2)

    def test_merge_unique_duplicate2_instance(self):
        c = Classes(TESTLIST1)
        c.merge_unique(Classes(TESTLIST2 + TESTLIST2))
        self.assertSequenceEqual(c, TESTLIST1 + TESTLIST2)

    def test_merge_unique_nonstring(self):
        c = Classes()
        with self.assertRaises(TypeError):
            c.merge_unique([0,1,2])

    def test_repr_empty(self):
        c = Classes()
        self.assertEqual('%r' % c, '%s(%r)' % (c.__class__.__name__, []))

    def test_repr_contents(self):
        c = Classes(TESTLIST1)
        self.assertEqual('%r' % c, '%s(%r)' % (c.__class__.__name__, TESTLIST1))

    def test_as_list(self):
        c = Classes(TESTLIST1)
        self.assertListEqual(c.as_list(), TESTLIST1)

if __name__ == '__main__':
    unittest.main()



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import sys

from reclass.errors import ChangedConstantError, ResolveError, TypeMergeError


class ValueList(object):

    def __init__(self, value, settings):
        self._settings = settings
        self._refs = []
        self.allRefs = True
        self._values = [value]
        self._inv_refs = []
        self.has_inv_query = False
        self.ignore_failed_render = False
        self.is_complex = False
        self._update()

    @property
    def uri(self):
        return '; '.join([str(x.uri) for x in self._values])

    def append(self, value):
        self._values.append(value)
        self._update()

    def extend(self, values):
        self._values.extend(values._values)
        self._update()

    def _update(self):
        self.assembleRefs()
        self._check_for_inv_query()
        self.is_complex = False
        item_type = self._values[0].item_type()
        for v in self._values:
            if v.is_complex or v.constant or v.overwrite or v.item_type() != item_type:
                self.is_complex = True

    @property
    def has_references(self):
        return len(self._refs) > 0

    def get_inv_references(self):
        return self._inv_refs

    def get_references(self):
        return self._refs

    def _check_for_inv_query(self):
        self.has_inv_query = False
        self.ignore_failed_render = True
        for value in self._values:
            if value.has_inv_query:
                self._inv_refs.extend(value.get_inv_references())
                self.has_inv_query = True
                if value.ignore_failed_render() is False:
                    self.ignore_failed_render = False
        if self.has_inv_query is False:
            self.ignore_failed_render = False

    def assembleRefs(self, context={}):
        self._refs = []
        self.allRefs = True
        for value in self._values:
            value.assembleRefs(context)
            if value.has_references:
                self._refs.extend(value.get_references())
            if value.allRefs is False:
                self.allRefs = False

    @property
    def needs_all_envs(self):
        for value in self._values:
            if value.needs_all_envs:
                return True
        return False

    def merge(self):
        output = None
        for n, value in enumerate(self._values):
            if output is None:
                output = value
            else:
                output = value.merge_over(output)
        return output

    def render(self, context, inventory):
        from reclass.datatypes.parameters import Parameters

        output = None
        deepCopied = False
        last_error = None
        constant = False
        for n, value in enumerate(self._values):
            try:
                new = value.render(context, inventory)
            except ResolveError as e:
                # only ignore failed renders if
                # ignore_overwritten_missing_references is set and we are
                # dealing with a scalar value and it's not the last item in the
                # values list
                if (self._settings.ignore_overwritten_missing_references
                        and not isinstance(output, (dict, list))
                        and n != (len(self._values)-1)):
                    new = None
                    last_error = e
                    print("[WARNING] Reference '%s' undefined" % str(value),
                          file=sys.stderr)
                else:
                    raise e

            if constant:
                if self._settings.strict_constant_parameters:
                    raise ChangedConstantError('{0}; {1}'.format(self._values[n-1].uri, self._values[n].uri))
                else:
                    continue

            if output is None or value.overwrite:
                output = new
                deepCopied = False
            else:
                if isinstance(output, dict):
                    if isinstance(new, dict):
                        p1 = Parameters(output, self._settings, None, parse_strings=False)
                        p2 = Parameters(new, self._settings, None, parse_strings=False)
                        p1.merge(p2)
                        output = p1.as_dict()
                    elif isinstance(new, list):
                        raise TypeMergeError(self._values[n], self._values[n-1], self.uri)
                    elif self._settings.allow_scalar_over_dict or (self._settings.allow_none_override and new is None):
                        output = new
                        deepCopied = False
                    else:
                        raise TypeMergeError(self._values[n], self._values[n-1], self.uri)
                elif isinstance(output, list):
                    if isinstance(new, list):
                        if not deepCopied:
                            output = copy.deepcopy(output)
                            deepCopied = True
                        output.extend(new)
                    elif isinstance(new, dict):
                        raise TypeMergeError(self._values[n], self._values[n-1], self.uri)
                    elif self._settings.allow_scalar_over_list or (self._settings.allow_none_override and new is None):
                        output = new
                        deepCopied = False
                    else:
                        raise TypeMergeError(self._values[n], self._values[n-1], self.uri)
                else:
                    if isinstance(new, dict):
                        if self._settings.allow_dict_over_scalar:
                            output = new
                            deepCopied = False
                        else:
                            raise TypeMergeError(self._values[n], self._values[n-1], self.uri)
                    elif isinstance(new, list):
                        if self._settings.allow_list_over_scalar:
                            output_list = list()
                            output_list.append(output)
                            output_list.extend(new)
                            output = output_list
                            deepCopied = True
                        else:
                            raise TypeMergeError(self._values[n], self._values[n-1], self.uri)
                    else:
                        output = new
                        deepCopied = False

            if value.constant:
                constant = True

        if isinstance(output, (dict, list)) and last_error is not None:
            raise last_error

        return output



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import copy

from six import iteritems

import reclass.errors
from reclass import get_storage
from reclass.storage import ExternalNodeStorageBase

def path_mangler(inventory_base_uri, nodes_uri, classes_uri):
    if nodes_uri == classes_uri:
        raise errors.DuplicateUriError(nodes_uri, classes_uri)
    return nodes_uri, classes_uri

STORAGE_NAME = 'mixed'

class ExternalNodeStorage(ExternalNodeStorageBase):

    MixedUri = collections.namedtuple('MixedURI', 'storage_type options')

    def __init__(self, nodes_uri, classes_uri, compose_node_name):
        super(ExternalNodeStorage, self).__init__(STORAGE_NAME, compose_node_name)

        self._nodes_uri = self._uri(nodes_uri)
        self._nodes_storage = get_storage(self._nodes_uri.storage_type, self._nodes_uri.options, None, compose_node_name)
        self._classes_default_uri = self._uri(classes_uri)
        self._classes_default_storage = get_storage(self._classes_default_uri.storage_type, None, self._classes_default_uri.options, compose_node_name)

        self._classes_storage = dict()
        if 'env_overrides' in classes_uri:
            for override in classes_uri['env_overrides']:
                for (env, options) in iteritems(override):
                        uri = copy.deepcopy(classes_uri)
                        uri.update(options)
                        uri = self._uri(uri)
                        self._classes_storage[env] = get_storage(uri.storage_type, None, uri.options, compose_node_name)

    def _uri(self, uri):
        ret = copy.deepcopy(uri)
        ret['storage_type'] = uri['storage_type']
        if 'env_overrides' in ret:
            del ret['env_overrides']
        if uri['storage_type'] == 'yaml_fs':
            ret = ret['uri']
        return self.MixedUri(uri['storage_type'], ret)

    def get_node(self, name, settings):
        return self._nodes_storage.get_node(name, settings)

    def get_class(self, name, environment, settings):
        storage = self._classes_storage.get(environment, self._classes_default_storage)
        return storage.get_class(name, environment, settings)

    def enumerate_nodes(self):
        return self._nodes_storage.enumerate_nodes()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

"""
Kapitan setup.py for PIP install
"""

from setuptools import setup, find_packages
from kapitan.version import PROJECT_NAME, VERSION, DESCRIPTION, URL, AUTHOR, AUTHOR_EMAIL, LICENCE


# From https://github.com/pypa/pip/issues/3610#issuecomment-356687173
def install_deps():
    """Reads requirements.txt and preprocess it
    to be feed into setuptools.

    This is the only possible way (we found)
    how requirements.txt can be reused in setup.py
    using dependencies from private github repositories.

    Links must be appendend by `-{StringWithAtLeastOneNumber}`
    or something like that, so e.g. `-9231` works as well as
    `1.1.0`. This is ignored by the setuptools, but has to be there.

    Returns:
         list of packages and dependency links.
    """
    with open("requirements.txt", "r") as f:
        packages = f.readlines()
        new_pkgs = []
        links = []
        for resource in packages:
            if "git+https" in resource:
                pkg = resource.split("#")[-1]
                links.append(resource.strip() + "-9876543210")
                new_pkgs.append(pkg.replace("egg=", "").rstrip())
            else:
                new_pkgs.append(resource.strip())
        return new_pkgs, links


pkgs, new_links = install_deps()

setup(
    name=PROJECT_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=URL,
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENCE,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="jsonnet kubernetes reclass jinja",
    py_modules=["kapitan"],
    python_requires=">=3.6",
    packages=find_packages(),
    include_package_data=True,
    dependency_links=new_links,
    install_requires=pkgs,
    entry_points={"console_scripts": ["kapitan=kapitan.cli:main",],},
)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import itertools as it
import operator
import pyparsing as pp

from six import iteritems
from six import string_types

from reclass.values import item
from reclass.values import parser_funcs
from reclass.settings import Settings
from reclass.utils.dictpath import DictPath
from reclass.errors import ExpressionError, ParseError, ResolveError


# TODO: generalize expression handling.
class BaseTestExpression(object):

    known_operators = {}
    def __init__(self, delimiter):
        self._delimiter = delimiter
        self.refs = []
        self.inv_refs = []


class EqualityTest(BaseTestExpression):

    known_operators = { parser_funcs.EQUAL: operator.eq,
                        parser_funcs.NOT_EQUAL: operator.ne}

    def __init__(self, expression, delimiter):
        # expression is a list of at least three tuples, of which first element
        # is a string tag, second is subelement value; other tuples apparently
        # are not used.
        # expression[0][1] effectively contains export path and apparently must
        # be treated as such, also left hand operand in comparison
        # expression[1][1] appa holds commparison operator == or !=
        # expression[2][1] is the righhand operand
        super(EqualityTest, self).__init__(delimiter)
        # TODO: this double sommersault must be cleaned
        _ = self._get_vars(expression[2][1], *self._get_vars(expression[0][1]))
        self._export_path, self._parameter_path, self._parameter_value = _
        try:
            self._export_path.drop_first()
        except AttributeError:
            raise ExpressionError('No export')
        try:
            self._compare = self.known_operators[expression[1][1]]
        except KeyError as e:
            msg = 'Unknown test {0}'.format(expression[1][1])
            raise ExpressionError(msg, tbFlag=False)
        self.inv_refs = [self._export_path]
        if self._parameter_path is not None:
            self._parameter_path.drop_first()
            self.refs = [str(self._parameter_path)]

    def value(self, context, items):
        if self._parameter_path is not None:
            self._parameter_value = self._resolve(self._parameter_path,
                                                  context)
        if self._parameter_value is None:
            raise ExpressionError('Failed to render %s' % str(self),
                                  tbFlag=False)
        if self._export_path.exists_in(items):
            export_value = self._resolve(self._export_path, items)
            return self._compare(export_value, self._parameter_value)
        return False

    def _resolve(self, path, dictionary):
        try:
            return path.get_value(dictionary)
        except KeyError as e:
            raise ResolveError(str(path))

    def _get_vars(self, var, export=None, parameter=None, value=None):
        if isinstance(var, string_types):
            path = DictPath(self._delimiter, var)
            if path.path[0].lower() == 'exports':
                export = path
            elif path.path[0].lower() == 'self':
                parameter = path
            elif path.path[0].lower() == 'true':
                value = True
            elif path.path[0].lower() == 'false':
                value = False
            else:
                value = var
        else:
            value = var
        return export, parameter, value


class LogicTest(BaseTestExpression):

    known_operators = { parser_funcs.AND: operator.and_,
                        parser_funcs.OR: operator.or_}

    def __init__(self, expr, delimiter):
        super(LogicTest, self).__init__(delimiter)
        subtests = list(it.compress(expr, it.cycle([1, 1, 1, 0])))
        self._els = [EqualityTest(subtests[j:j+3], self._delimiter)
                     for j in range(0, len(subtests), 3)]
        for x in self._els:
            self.refs.extend(x.refs)
            self.inv_refs.extend(x.inv_refs)
        try:
            self._ops = [self.known_operators[x[1]] for x in expr[3::4]]
        except KeyError as e:
            msg = 'Unknown operator {0} {1}'.format(e.messsage, self._els)
            raise ExpressionError(msg, tbFlag=False)

    def value(self, context, items):
        if len(self._els) == 0:  # NOTE: possible logic error
            return True
        result = self._els[0].value(context, items)
        for op, next_el in zip(self._ops, self._els[1:]):
            result = op(result, next_el.value(context, items))
        return result


class InvItem(item.Item):

    type = item.ItemTypes.INV_QUERY

    def __init__(self, newitem, settings):
        super(InvItem, self).__init__(newitem.render(None, None), settings)
        self.needs_all_envs = False
        self.has_inv_query = True
        self.ignore_failed_render = (
                self._settings.inventory_ignore_failed_render)
        self._parse_expression(self.contents)

    def _parse_expression(self, expr):
        parser = parser_funcs.get_expression_parser()
        try:
            tokens = parser.parseString(expr).asList()
        except pp.ParseException as e:
            raise ParseError(e.msg, e.line, e.col, e.lineno)

        if len(tokens) == 2:  # options are set
            passed_opts = [x[1] for x in tokens.pop(0)]
            self.ignore_failed_render = parser_funcs.IGNORE_ERRORS in passed_opts
            self.needs_all_envs = parser_funcs.ALL_ENVS in passed_opts
        elif len(tokens) > 2:
            raise ExpressionError('Failed to parse %s' % str(tokens),
                                  tbFlag=False)
        self._expr_type = tokens[0][0]
        self._expr = list(tokens[0][1])

        if self._expr_type == parser_funcs.VALUE:
            self._value_path = DictPath(self._settings.delimiter,
                                        self._expr[0][1]).drop_first()
            self._question = LogicTest([], self._settings.delimiter)
            self.refs = []
            self.inv_refs = [self._value_path]
        elif self._expr_type == parser_funcs.TEST:
            self._value_path = DictPath(self._settings.delimiter,
                                        self._expr[0][1]).drop_first()
            self._question = LogicTest(self._expr[2:], self._settings.delimiter)
            self.refs = self._question.refs
            self.inv_refs = self._question.inv_refs
            self.inv_refs.append(self._value_path)
        elif self._expr_type == parser_funcs.LIST_TEST:
            self._value_path = None
            self._question = LogicTest(self._expr[1:], self._settings.delimiter)
            self.refs = self._question.refs
            self.inv_refs = self._question.inv_refs
        else:
            msg = 'Unknown expression type: %s'
            raise ExpressionError(msg % self._expr_type, tbFlag=False)

    @property
    def has_references(self):
        return len(self._question.refs) > 0

    def get_references(self):
        return self._question.refs

    def assembleRefs(self, context):
        return

    def get_inv_references(self):
        return self.inv_refs

    def _resolve(self, path, dictionary):
        try:
            return path.get_value(dictionary)
        except KeyError as e:
            raise ResolveError(str(path))

    def _value_expression(self, inventory):
        results = {}
        for (node, items) in iteritems(inventory):
            if self._value_path.exists_in(items):
                results[node] = copy.deepcopy(self._resolve(self._value_path,
                                              items))
        return results

    def _test_expression(self, context, inventory):
        if self._value_path is None:
            msg = 'Failed to render %s'
            raise ExpressionError(msg % str(self), tbFlag=False)

        results = {}
        for node, items in iteritems(inventory):
            if (self._question.value(context, items) and
                    self._value_path.exists_in(items)):
                results[node] = copy.deepcopy(
                    self._resolve(self._value_path, items))
        return results

    def _list_test_expression(self, context, inventory):
        results = []
        for (node, items) in iteritems(inventory):
            if self._question.value(context, items):
                results.append(node)
        return results

    def render(self, context, inventory):
        if self._expr_type == parser_funcs.VALUE:
            return self._value_expression(inventory)
        elif self._expr_type == parser_funcs.TEST:
            return self._test_expression(context, inventory)
        elif self._expr_type == parser_funcs.LIST_TEST:
            return self._list_test_expression(context, inventory)
        raise ExpressionError('Failed to render %s' % str(self), tbFlag=False)

    def __str__(self):
        return ' '.join(str(j) for i,j in self._expr)

    def __repr__(self):
        # had to leave it here for now as the behaviour differs from basic
        return 'InvItem(%r)' % self._expr



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

"jsonnet tests"

import json
import os
import unittest

from kapitan.resources import (
    gzip_b64,
    jsonschema_validate,
    yaml_dump,
    yaml_dump_stream,
    yaml_load,
    yaml_load_stream,
)
from kapitan.utils import prune_empty, sha256_string


class JsonnetNativeFuncsTest(unittest.TestCase):
    def test_yaml_dump(self):
        """dump json string to yaml"""
        yaml = yaml_dump('{"key":"value"}')
        self.assertEqual(yaml, "key: value\n")

    def test_yaml_dump_stream(self):
        """dump json string to yaml"""
        yaml = yaml_dump_stream('[{"key":"value"},{"key":"value"}]')
        self.assertEqual(yaml, "key: value\n---\nkey: value\n")

    def test_yaml_load(self):
        """
            This tests the yaml_load function.
            It converts the yaml file in test_resources/ to a json string
        """
        current_pwd = os.path.dirname(__file__)
        json = yaml_load([current_pwd], "test_resources/test_yaml_load.yaml")
        expected_output = """{"test": {"key": "value", "array": ["ele1", "ele2"]}}"""
        self.assertEqual(json, expected_output)

    def test_yaml_load_stream(self):
        """
            This tests the yaml_load_stream function.
            It converts the yaml file in test_resources/ to a json string
        """
        current_pwd = os.path.dirname(__file__)
        json = yaml_load_stream([current_pwd], "test_resources/test_yaml_load_stream.yaml")
        expected_output = """[{"test1": {"key": "value", "array": ["ele1", "ele2"]}}, {"test2": {"key": "value", "array": ["ele1", "ele2"]}}]"""
        self.assertEqual(json, expected_output)

    def test_sha256_string(self):
        """sha256 hex digest for string"""
        hash = sha256_string("test")
        self.assertEqual(hash, "9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08")

    def test_gzip_b64(self):
        """base64-encoded gzip-compression for string"""
        gzip = gzip_b64("test")
        self.assertEqual(gzip, "H4sIAAAAAAAC/ytJLS4BAAx+f9gEAAAA")

    def test_prune_empty(self):
        """Remove empty lists and empty dictionaries from dict"""
        dictionary = {"hello": "world", "array": [1, 2], "foo": {}, "bar": []}
        pruned = prune_empty(dictionary)
        self.assertEqual(pruned, {"hello": "world", "array": [1, 2]})

    def test_jsonschema_valid(self):
        """validate valid obj with jsonschema"""
        dictionary = {"msg": "hello, world!", "array": [1, 2]}
        schema = {
            "type": "object",
            "properties": {
                "msg": {"type": "string"},
                "array": {"type": "array", "contains": {"type": "number"}},
            },
        }
        validation = jsonschema_validate(json.dumps(dictionary), json.dumps(schema))

        self.assertTrue(validation["valid"])
        self.assertEqual(validation["reason"], "")

    def test_jsonschema_invalid(self):
        """validate invalid obj with jsonschema"""
        dictionary = {"msg": "hello, world!", "array": ["a", "b", "c"]}
        schema = {
            "type": "object",
            "properties": {
                "msg": {"type": "string"},
                "array": {"type": "array", "contains": {"type": "number"}},
            },
        }
        validation = jsonschema_validate(json.dumps(dictionary), json.dumps(schema))

        self.assertFalse(validation["valid"])
        self.assertNotEqual(validation["reason"], "")



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

"base64 ref module"

import base64
import errno
import json
import logging

import yaml
from kapitan.refs.base import PlainRef, PlainRefBackend

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader

logger = logging.getLogger(__name__)


class Base64Ref(PlainRef):
    def __init__(self, data, from_base64=False, **kwargs):
        """
        writes data
        set from_base64 to load already base64 encoded data
        """
        super().__init__(data, **kwargs)
        self.type_name = "base64"
        self.encoding = kwargs.get("encoding", "original")
        self.embed_refs = kwargs.get("embed_refs", False)

        # TODO data should be bytes only
        if from_base64:
            self.data = data
        else:
            self.data = base64.b64encode(data).decode()

    def reveal(self):
        # TODO data should be bytes only
        return base64.b64decode(self.data).decode()

    def compile_embedded(self):
        dump = self.dump()
        # if subvar is set, save path in 'embedded_subvar_path' key
        subvar = self.path.split("@")
        if len(subvar) > 1:
            dump["embedded_subvar_path"] = subvar[1]
        dump_data = base64.b64encode(json.dumps(dump).encode()).decode()
        return f"?{{{self.type_name}:{dump_data}:embedded}}"

    def compile(self):
        # XXX will only work if object read via backend

        if self.embed_refs:
            return self.compile_embedded()

        return f"?{{{self.type_name}:{self.path}:{self.hash[:8]}}}"

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        """
        return a new Base64Ref from file at ref_full_path
        the data key in the file must be base64 encoded
        """
        try:
            with open(ref_full_path) as fp:
                obj = yaml.load(fp, Loader=YamlLoader)
                _kwargs = {key: value for key, value in obj.items() if key not in ("data", "from_base64")}
                kwargs.update(_kwargs)
                return cls(obj["data"], from_base64=True, **kwargs)

        except IOError as ex:
            if ex.errno == errno.ENOENT:
                return None

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {"data": self.data, "encoding": self.encoding, "type": self.type_name}


class Base64RefBackend(PlainRefBackend):
    def __init__(self, path, ref_type=Base64Ref, **ref_kwargs):
        super().__init__(path, ref_type, **ref_kwargs)
        self.type_name = "base64"



# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
from functools import lru_cache

import jsonschema
import requests
import yaml
from kapitan import defaults
from kapitan.errors import KubernetesManifestValidationError, RequestUnsuccessfulError
from kapitan.utils import make_request
from kapitan.validator.base import Validator

logger = logging.getLogger(__name__)


class KubernetesManifestValidator(Validator):
    def __init__(self, cache_dir, **kwargs):
        super().__init__(cache_dir, **kwargs)

    def validate(self, validate_paths, **kwargs):
        """
        validates manifests at validate_paths against json schema as specified by
        'kind' and 'version' in kwargs.
        raises KubernetesManifestValidationError encountering the first validation error
        """
        kind = kwargs.get("kind")
        version = kwargs.get("version", defaults.DEFAULT_KUBERNETES_VERSION)
        schema = self._get_schema(kind, version)
        validator = jsonschema.Draft4Validator(schema)
        for validate_path in validate_paths:
            if not os.path.isfile(validate_path):
                logger.warning("{} does not exist. skipping".format(validate_path))
                continue
            with open(validate_path, "r") as fp:
                validate_instance = yaml.safe_load(fp.read())
                errors = sorted(validator.iter_errors(validate_instance), key=lambda e: e.path)
                if errors:
                    error_message = "invalid '{}' manifest at {}\n".format(kind, validate_path)
                    error_message += "\n".join(
                        ["{} {}".format(list(error.path), error.message) for error in errors]
                    )
                    raise KubernetesManifestValidationError(error_message)
                else:
                    logger.info("Validation: manifest validation successful for {}".format(validate_path))

    @lru_cache(maxsize=256)
    def _get_schema(self, kind, version):
        """gets json validation schema from web or cache"""
        schema = self._get_cached_schema(kind, version)
        if schema is None:
            schema = self._get_schema_from_web(kind, version)
            self._cache_schema(kind, version, schema)
        return schema

    def _cache_schema(self, kind, version, schema):
        cache_path = self._get_cache_path(kind, version)
        os.makedirs(os.path.dirname(cache_path), exist_ok=True)
        with open(cache_path, "w") as fp:
            yaml.safe_dump(schema, stream=fp, default_flow_style=False)

    def _get_cached_schema(self, kind, version):
        cache_path = self._get_cache_path(kind, version)
        if os.path.isfile(cache_path):
            with open(cache_path, "r") as fp:
                return yaml.safe_load(fp.read())
        else:
            return None

    def _get_schema_from_web(self, kind, version):
        url = self._get_request_url(kind, version)
        logger.debug("Validation: fetching schema from {}".format(url))
        try:
            content, _ = make_request(url)
        except requests.exceptions.HTTPError:
            raise RequestUnsuccessfulError(
                "Validation: schema failed to fetch from {}"
                "\nThe specified version '{}' or kind '{}' may not be supported".format(url, version, kind)
            )
        logger.debug("Validation: schema fetched  from {}".format(url))
        return yaml.safe_load(content)

    def _get_request_url(self, kind, version):
        return "https://kubernetesjsonschema.dev/" + defaults.FILE_PATH_FORMAT.format(version, kind)

    def _get_cache_path(self, kind, version):
        return os.path.join(self.cache_dir, defaults.FILE_PATH_FORMAT.format(version, kind))



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

"version check tests"

import unittest

from kapitan.utils import compare_versions


class VersionTest(unittest.TestCase):
    "Test version checks"

    def test_version_equal(self):
        dot_kapitan_version = "0.22.0"
        current_version = "0.22.0"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_rc_equal(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_major_equal(self):
        dot_kapitan_version = "0.22"
        current_version = "0.22.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_one_rc_major_equal(self):
        dot_kapitan_version = "0.22"
        current_version = "0.22.1-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_greater(self):
        dot_kapitan_version = "0.22.1"
        current_version = "0.22.0"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_one_rc_greater(self):
        dot_kapitan_version = "0.22.0"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_both_rc_greater(self):
        dot_kapitan_version = "0.22.1-rc.1"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_both_rc_major_greater(self):
        dot_kapitan_version = "0.23.0-rc.1"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_lower(self):
        dot_kapitan_version = "0.22.0"
        current_version = "0.22.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")

    def test_version_one_rc_lower(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.22.0"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")

    def test_version_both_rc_lower(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.22.1-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")

    def test_version_both_rc_major_lower(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.23.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")


#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
import os
from reclass.errors import InvalidClassnameError

INVALID_CHARACTERS_FOR_CLASSNAMES = ' ' + os.sep


class Classes(object):
    '''
    A very limited ordered set of strings with O(n) uniqueness constraints. It
    is neither a proper list or a proper set, on purpose, to keep things
    simple.
    '''
    def __init__(self, iterable=None):
        self._items = []
        if iterable is not None:
            self.merge_unique(iterable)

    def __len__(self):
        return len(self._items)

    def __eq__(self, rhs):
        if isinstance(rhs, list):
            return self._items == rhs
        else:
            try:
                return self._items == rhs._items
            except AttributeError as e:
                return False

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def as_list(self):
        return self._items[:]

    def merge_unique(self, iterable):
        if isinstance(iterable, self.__class__):
            iterable = iterable.as_list()
        # Cannot just call list.extend here, as iterable's items might not
        # be unique by themselves, or in the context of self.
        for i in iterable:
            self.append_if_new(i)

    def _assert_is_string(self, item):
        if not isinstance(item, six.string_types):
            raise TypeError('%s instances can only contain strings, '
                            'not %s' % (self.__class__.__name__, type(item)))

    def _assert_valid_characters(self, item):
        for c in INVALID_CHARACTERS_FOR_CLASSNAMES:
            if c in item:
                raise InvalidClassnameError(c, item)

    def _append_if_new(self, item):
        if item not in self._items:
            self._items.append(item)

    def append_if_new(self, item):
        self._assert_is_string(item)
        self._assert_valid_characters(item)
        self._append_if_new(item)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._items)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy

from six import iteritems

from reclass.settings import Settings
from reclass.datatypes import Parameters
from reclass.utils.parameterdict import ParameterDict
from reclass.utils.parameterlist import ParameterList
from reclass.values.value import Value
from reclass.values.valuelist import ValueList
from reclass.values.scaitem import ScaItem
from reclass.errors import ChangedConstantError, InfiniteRecursionError, InterpolationError, ResolveError, ResolveErrorList, TypeMergeError
import unittest

try:
    import unittest.mock as mock
except ImportError:
    import mock

SIMPLE = {'one': 1, 'two': 2, 'three': 3}
SETTINGS = Settings()

class MockDevice(object):
    def __init__(self):
        self._text = ''

    def write(self, s):
        self._text += s
        return

    def text(self):
        return self._text

class TestParameters(unittest.TestCase):

    def _construct_mocked_params(self, iterable=None, settings=SETTINGS):
        p = Parameters(iterable, settings, '')
        self._base = base = p._base
        p._base = mock.MagicMock(spec_set=ParameterDict, wraps=base)
        p._base.__repr__ = mock.MagicMock(autospec=dict.__repr__,
                                          return_value=repr(base))
        p._base.__getitem__.side_effect = base.__getitem__
        p._base.__setitem__.side_effect = base.__setitem__
        return p, p._base

    def test_len_empty(self):
        p, b = self._construct_mocked_params()
        l = 0
        b.__len__.return_value = l
        self.assertEqual(len(p), l)
        b.__len__.assert_called_with()

    def test_constructor(self):
        p, b = self._construct_mocked_params(SIMPLE)
        l = len(SIMPLE)
        b.__len__.return_value = l
        self.assertEqual(len(p), l)
        b.__len__.assert_called_with()

    def test_repr_empty(self):
        p, b = self._construct_mocked_params()
        b.__repr__.return_value = repr({})
        self.assertEqual('%r' % p, '%s(%r)' % (p.__class__.__name__, {}))
        b.__repr__.assert_called_once_with()

    def test_repr(self):
        p, b = self._construct_mocked_params(SIMPLE)
        b.__repr__.return_value = repr(SIMPLE)
        self.assertEqual('%r' % p, '%s(%r)' % (p.__class__.__name__, SIMPLE))
        b.__repr__.assert_called_once_with()

    def test_equal_empty(self):
        p1, b1 = self._construct_mocked_params()
        p2, b2 = self._construct_mocked_params()
        b1.__eq__.return_value = True
        self.assertEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_equal_default_delimiter(self):
        p1, b1 = self._construct_mocked_params(SIMPLE)
        p2, b2 = self._construct_mocked_params(SIMPLE, SETTINGS)
        b1.__eq__.return_value = True
        self.assertEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_equal_contents(self):
        p1, b1 = self._construct_mocked_params(SIMPLE)
        p2, b2 = self._construct_mocked_params(SIMPLE)
        b1.__eq__.return_value = True
        self.assertEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_unequal_content(self):
        p1, b1 = self._construct_mocked_params()
        p2, b2 = self._construct_mocked_params(SIMPLE)
        b1.__eq__.return_value = False
        self.assertNotEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_unequal_delimiter(self):
        settings1 = Settings({'delimiter': ':'})
        settings2 = Settings({'delimiter': '%'})
        p1, b1 = self._construct_mocked_params(settings=settings1)
        p2, b2 = self._construct_mocked_params(settings=settings2)
        b1.__eq__.return_value = False
        self.assertNotEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_unequal_types(self):
        p1, b1 = self._construct_mocked_params()
        self.assertNotEqual(p1, None)
        self.assertEqual(b1.__eq__.call_count, 0)

    def test_construct_wrong_type(self):
        with self.assertRaises(TypeError) as e:
            self._construct_mocked_params(str('wrong type'))
        self.assertIn(str(e.exception), [ "Cannot merge <type 'str'> objects into Parameters",    # python 2
                                          "Cannot merge <class 'str'> objects into Parameters" ])  # python 3

    def test_merge_wrong_type(self):
        p, b = self._construct_mocked_params()
        with self.assertRaises(TypeError) as e:
            p.merge(str('wrong type'))
        self.assertIn(str(e.exception), [ "Cannot merge <type 'str'> objects into Parameters",    # python 2
                                          "Cannot merge <class 'str'> objects into Parameters"])   # python 3

    def test_get_dict(self):
        p, b = self._construct_mocked_params(SIMPLE)
        p.initialise_interpolation()
        self.assertDictEqual(p.as_dict(), SIMPLE)

    def test_merge_scalars(self):
        p1, b1 = self._construct_mocked_params(SIMPLE)
        mergee = {'five':5,'four':4,'None':None,'tuple':(1,2,3)}
        p2, b2 = self._construct_mocked_params(mergee)
        p1.merge(p2)
        self.assertEqual(b1.get.call_count, 4)
        self.assertEqual(b1.__setitem__.call_count, 4)

    def test_stray_occurrence_overwrites_during_interpolation(self):
        p1 = Parameters({'r' : mock.sentinel.ref, 'b': '${r}'}, SETTINGS, '')
        p2 = Parameters({'b' : mock.sentinel.goal}, SETTINGS, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict()['b'], mock.sentinel.goal)


class TestParametersNoMock(unittest.TestCase):

    def test_merge_scalars(self):
        p = Parameters(SIMPLE, SETTINGS, '')
        mergee = {'five':5,'four':4,'None':None,'tuple':(1,2,3)}
        p.merge(mergee)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), goal)

    def test_merge_scalars_overwrite(self):
        p = Parameters(SIMPLE, SETTINGS, '')
        mergee = {'two':5,'four':4,'three':None,'one':(1,2,3)}
        p.merge(mergee)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), goal)

    def test_merge_lists(self):
        l1 = [1,2,3]
        l2 = [2,3,4]
        p1 = Parameters(dict(list=l1[:]), SETTINGS, '')
        p2 = Parameters(dict(list=l2), SETTINGS, '')
        p1.merge(p2)
        p1.initialise_interpolation()
        self.assertListEqual(p1.as_dict()['list'], l1+l2)

    def test_merge_list_into_scalar(self):
        l = ['foo', 1, 2]
        p1 = Parameters(dict(key=l[0]), SETTINGS, '')
        p2 = Parameters(dict(key=l[1:]), SETTINGS, '')
        with self.assertRaises(TypeMergeError) as e:
            p1.merge(p2)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Cannot merge list over scalar, at key, in ; ")

    def test_merge_list_into_scalar_allow(self):
        settings = Settings({'allow_list_over_scalar': True})
        l = ['foo', 1, 2]
        p1 = Parameters(dict(key=l[0]), settings, '')
        p2 = Parameters(dict(key=l[1:]), settings, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertListEqual(p1.as_dict()['key'], l)

    def test_merge_scalar_over_list(self):
        l = ['foo', 1, 2]
        p1 = Parameters(dict(key=l[:2]), SETTINGS, '')
        p2 = Parameters(dict(key=l[2]), SETTINGS, '')
        with self.assertRaises(TypeMergeError) as e:
            p1.merge(p2)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Cannot merge scalar over list, at key, in ; ")

    def test_merge_scalar_over_list_allow(self):
        l = ['foo', 1, 2]
        settings = Settings({'allow_scalar_over_list': True})
        p1 = Parameters(dict(key=l[:2]), settings, '')
        p2 = Parameters(dict(key=l[2]), settings, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict()['key'], l[2])

    def test_merge_none_over_list(self):
        l = ['foo', 1, 2]
        settings = Settings({'allow_none_override': False})
        p1 = Parameters(dict(key=l[:2]), settings, '')
        p2 = Parameters(dict(key=None), settings, '')
        with self.assertRaises(TypeMergeError) as e:
            p1.merge(p2)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Cannot merge scalar over list, at key, in ; ")

    def test_merge_none_over_list_allow(self):
        l = ['foo', 1, 2]
        settings = Settings({'allow_none_override': True})
        p1 = Parameters(dict(key=l[:2]), settings, '')
        p2 = Parameters(dict(key=None), settings, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict()['key'], None)

    def test_merge_dict_over_scalar(self):
        d = { 'one': 1, 'two': 2 }
        p1 = Parameters({ 'a': 1 }, SETTINGS, '')
        p2 = Parameters({ 'a': d }, SETTINGS, '')
        with self.assertRaises(TypeMergeError) as e:
            p1.merge(p2)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Cannot merge dictionary over scalar, at a, in ; ")

    def test_merge_dict_over_scalar_allow(self):
        settings = Settings({'allow_dict_over_scalar': True})
        d = { 'one': 1, 'two': 2 }
        p1 = Parameters({ 'a': 1 }, settings, '')
        p2 = Parameters({ 'a': d }, settings, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), { 'a': d })

    def test_merge_scalar_over_dict(self):
        d = { 'one': 1, 'two': 2}
        p1 = Parameters({ 'a': d }, SETTINGS, '')
        p2 = Parameters({ 'a': 1 }, SETTINGS, '')
        with self.assertRaises(TypeMergeError) as e:
            p1.merge(p2)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Cannot merge scalar over dictionary, at a, in ; ")

    def test_merge_scalar_over_dict_allow(self):
        d = { 'one': 1, 'two': 2}
        settings = Settings({'allow_scalar_over_dict': True})
        p1 = Parameters({ 'a': d }, settings, '')
        p2 = Parameters({ 'a': 1 }, settings, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), { 'a': 1})

    def test_merge_none_over_dict(self):
        p1 = Parameters(dict(key=SIMPLE), SETTINGS, '')
        p2 = Parameters(dict(key=None), SETTINGS, '')
        with self.assertRaises(TypeMergeError) as e:
            p1.merge(p2)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Cannot merge scalar over dictionary, at key, in ; ")

    def test_merge_none_over_dict_allow(self):
        settings = Settings({'allow_none_override': True})
        p1 = Parameters(dict(key=SIMPLE), settings, '')
        p2 = Parameters(dict(key=None), settings, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict()['key'], None)

    def test_merge_list_over_dict(self):
        p1 = Parameters({}, SETTINGS, '')
        p2 = Parameters({'one': { 'a': { 'b': 'c' } } }, SETTINGS, 'second')
        p3 = Parameters({'one': { 'a': [ 'b' ] } }, SETTINGS, 'third')
        with self.assertRaises(TypeMergeError) as e:
            p1.merge(p2)
            p1.merge(p3)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Cannot merge list over dictionary, at one:a, in second; third")

    # def test_merge_bare_dict_over_dict(self):
        # settings = Settings({'allow_bare_override': True})
        # p1 = Parameters(dict(key=SIMPLE), settings, '')
        # p2 = Parameters(dict(key=dict()), settings, '')
        # p1.merge(p2)
        # p1.initialise_interpolation()
        # self.assertEqual(p1.as_dict()['key'], {})

    # def test_merge_bare_list_over_list(self):
        # l = ['foo', 1, 2]
        # settings = Settings({'allow_bare_override': True})
        # p1 = Parameters(dict(key=l), settings, '')
        # p2 = Parameters(dict(key=list()), settings, '')
        # p1.merge(p2)
        # p1.initialise_interpolation()
        # self.assertEqual(p1.as_dict()['key'], [])

    def test_merge_dicts(self):
        mergee = {'five':5,'four':4,'None':None,'tuple':(1,2,3)}
        p = Parameters(dict(dict=SIMPLE), SETTINGS, '')
        p2 = Parameters(dict(dict=mergee), SETTINGS, '')
        p.merge(p2)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), dict(dict=goal))

    def test_merge_dicts_overwrite(self):
        mergee = {'two':5,'four':4,'three':None,'one':(1,2,3)}
        p = Parameters(dict(dict=SIMPLE), SETTINGS, '')
        p2 = Parameters(dict(dict=mergee), SETTINGS, '')
        p.merge(p2)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), dict(dict=goal))

    def test_merge_dicts_override(self):
        """Validate that tilde merge overrides function properly."""
        mergee = {'~one': {'a': 'alpha'},
                  '~two': ['gamma']}
        base = {'one': {'b': 'beta'},
                'two': ['delta']}
        goal = {'one': {'a': 'alpha'},
                'two': ['gamma']}
        p = Parameters(dict(dict=base), SETTINGS, '')
        p2 = Parameters(dict(dict=mergee), SETTINGS, '')
        p.merge(p2)
        p.interpolate()
        self.assertDictEqual(p.as_dict(), dict(dict=goal))

    def test_interpolate_single(self):
        v = 42
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': v}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_multiple(self):
        v = '42'
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels) + 'meep'.join(SETTINGS.reference_sentinels),
             'bar': v[0],
             'meep': v[1]}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_multilevel(self):
        v = 42
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'meep'.join(SETTINGS.reference_sentinels),
             'meep': v}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_list(self):
        l = [41, 42, 43]
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': l}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], l)

    def test_interpolate_infrecursion(self):
        v = 42
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'foo'.join(SETTINGS.reference_sentinels)}
        p = Parameters(d, SETTINGS, '')
        with self.assertRaises(InfiniteRecursionError) as e:
            p.interpolate()
        # interpolation can start with foo or bar
        self.assertIn(e.exception.message, [ "-> \n   Infinite recursion: ${foo}, at bar",
                                             "-> \n   Infinite recursion: ${bar}, at foo"])

    def test_nested_references(self):
        d = {'a': '${${z}}', 'b': 2, 'z': 'b'}
        r = {'a': 2, 'b': 2, 'z': 'b'}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict(), r)

    def test_nested_deep_references(self):
        d = {'one': { 'a': 1, 'b': '${one:${one:c}}', 'c': 'a' } }
        r = {'one': { 'a': 1, 'b': 1, 'c': 'a'} }
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict(), r)

    def test_stray_occurrence_overwrites_during_interpolation(self):
        p1 = Parameters({'r' : 1, 'b': '${r}'}, SETTINGS, '')
        p2 = Parameters({'b' : 2}, SETTINGS, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict()['b'], 2)

    def test_referenced_dict_deep_overwrite(self):
        p1 = Parameters({'alpha': {'one': {'a': 1, 'b': 2} } }, SETTINGS, '')
        p2 = Parameters({'beta': '${alpha}'}, SETTINGS, '')
        p3 = Parameters({'alpha': {'one': {'c': 3, 'd': 4} },
                         'beta':  {'one': {'a': 99} } }, SETTINGS, '')
        r = {'alpha': {'one': {'a':1, 'b': 2, 'c': 3, 'd':4} },
             'beta': {'one': {'a':99, 'b': 2, 'c': 3, 'd':4} } }
        p1.merge(p2)
        p1.merge(p3)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_complex_reference_overwriting(self):
        p1 = Parameters({'one': 'abc_123_${two}_${three}', 'two': 'XYZ', 'four': 4}, SETTINGS, '')
        p2 = Parameters({'one': 'QWERTY_${three}_${four}', 'three': '999'}, SETTINGS, '')
        r = {'one': 'QWERTY_999_4', 'two': 'XYZ', 'three': '999', 'four': 4}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_nested_reference_with_overwriting(self):
        p1 = Parameters({'one': {'a': 1, 'b': 2, 'z': 'a'},
                         'two': '${one:${one:z}}' }, SETTINGS, '')
        p2 = Parameters({'one': {'z': 'b'} }, SETTINGS, '')
        r = {'one': {'a': 1, 'b':2, 'z': 'b'}, 'two': 2}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_merge_referenced_lists(self):
        p1 = Parameters({'one': [ 1, 2, 3 ], 'two': [ 4, 5, 6 ], 'three': '${one}'}, SETTINGS, '')
        p2 = Parameters({'three': '${two}'}, SETTINGS, '')
        r = {'one': [ 1, 2, 3 ], 'two': [ 4, 5, 6], 'three': [ 1, 2, 3, 4, 5, 6 ]}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_merge_referenced_dicts(self):
        p1 = Parameters({'one': {'a': 1, 'b': 2}, 'two': {'c': 3, 'd': 4}, 'three': '${one}'}, SETTINGS, '')
        p2 = Parameters({'three': '${two}'}, SETTINGS, '')
        r = {'one': {'a': 1, 'b': 2}, 'two': {'c': 3, 'd': 4}, 'three': {'a': 1, 'b': 2, 'c': 3, 'd': 4}}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_deep_refs_in_referenced_dicts(self):
        p = Parameters({'A': '${C:a}', 'B': {'a': 1, 'b': 2}, 'C': '${B}'}, SETTINGS, '')
        r = {'A': 1, 'B': {'a': 1, 'b': 2}, 'C': {'a': 1, 'b': 2}}
        p.interpolate()
        self.assertEqual(p.as_dict(), r)

    def test_overwrite_none(self):
        p1 = Parameters({'A': None, 'B': None, 'C': None, 'D': None, 'E': None, 'F': None}, SETTINGS, '')
        p2 = Parameters({'A': 'abc', 'B': [1, 2, 3], 'C': {'a': 'aaa', 'b': 'bbb'}, 'D': '${A}', 'E': '${B}', 'F': '${C}'}, SETTINGS, '')
        r = {'A': 'abc', 'B': [1, 2, 3], 'C': {'a': 'aaa', 'b': 'bbb'}, 'D': 'abc', 'E': [1, 2, 3], 'F': {'a': 'aaa', 'b': 'bbb'}}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_overwrite_dict(self):
        p1 = Parameters({'a': { 'one': 1, 'two': 2 }}, SETTINGS, '')
        p2 = Parameters({'~a': { 'three': 3, 'four': 4 }}, SETTINGS, '')
        r = {'a': { 'three': 3, 'four': 4 }}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_overwrite_list(self):
        p1 = Parameters({'a': [1, 2]}, SETTINGS, '')
        p2 = Parameters({'~a': [3, 4]}, SETTINGS, '')
        r = {'a': [3, 4]}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_interpolate_escaping(self):
        v = 'bar'.join(SETTINGS.reference_sentinels)
        d = {'foo': SETTINGS.escape_character + 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'unused'}
        p = Parameters(d, SETTINGS, '')
        p.initialise_interpolation()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_double_escaping(self):
        v = SETTINGS.escape_character + 'meep'
        d = {'foo': SETTINGS.escape_character + SETTINGS.escape_character + 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'meep'}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_escaping_backwards_compatibility(self):
        """In all following cases, escaping should not happen and the escape character
        needs to be printed as-is, to ensure backwards compatibility to older versions."""
        v = ' '.join([
            # Escape character followed by unescapable character
            '1', SETTINGS.escape_character,
            # Escape character followed by escape character
            '2', SETTINGS.escape_character + SETTINGS.escape_character,
            # Escape character followed by interpolation end sentinel
            '3', SETTINGS.escape_character + SETTINGS.reference_sentinels[1],
            # Escape character at the end of the string
            '4', SETTINGS.escape_character
            ])
        d = {'foo': v}
        p = Parameters(d, SETTINGS, '')
        p.initialise_interpolation()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_escape_close_in_ref(self):
        p1 = Parameters({'one}': 1, 'two': '${one\\}}'}, SETTINGS, '')
        r = {'one}': 1, 'two': 1}
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_double_escape_in_ref(self):
        d = {'one\\': 1, 'two': '${one\\\\}'}
        p1 = Parameters(d, SETTINGS, '')
        r = {'one\\': 1, 'two': 1}
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_merging_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': 111 }}, SETTINGS, '')
        p2 = Parameters({ 'beta': {'two': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': {'two': 222 }}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1'}, SETTINGS, '')
        r1 = { 'alpha': { 'one': 111 }, 'beta': { 'two': 111 }, 'name': 'node1' }
        r2 = { 'alpha': { 'one': 111 }, 'beta': { 'two': 222 }, 'name': 'node2' }
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({'name': 'node2'}, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_list_merging_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': [1, 2] }}, SETTINGS, '')
        p2 = Parameters({ 'beta': {'two': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': {'two': [3] }}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1'}, SETTINGS, '')
        r1 = { 'alpha': { 'one': [1, 2] }, 'beta': { 'two': [1, 2] }, 'name': 'node1' }
        r2 = { 'alpha': { 'one': [1, 2] }, 'beta': { 'two': [1, 2, 3] }, 'name': 'node2' }
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({'name': 'node2'}, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_dict_merging_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': { 'a': 'aa', 'b': 'bb' }}}, SETTINGS, '')
        p2 = Parameters({ 'beta': {'two': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': {'two': {'c': 'cc' }}}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1'}, SETTINGS, '')
        r1 = { 'alpha': { 'one': {'a': 'aa', 'b': 'bb'} }, 'beta': { 'two': {'a': 'aa', 'b': 'bb'} }, 'name': 'node1' }
        r2 = { 'alpha': { 'one': {'a': 'aa', 'b': 'bb'} }, 'beta': { 'two': {'a': 'aa', 'b': 'bb', 'c': 'cc'} }, 'name': 'node2' }
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({'name': 'node2'}, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_list_merging_with_refs_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': [1, 2], 'two': [3, 4] }}, SETTINGS, '')
        p2 = Parameters({ 'beta': { 'three': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': { 'three': '${alpha:two}' }}, SETTINGS, '')
        p4 = Parameters({ 'beta': { 'three': '${alpha:one}' }}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1' }, SETTINGS, '')
        r1 = {'alpha': {'one': [1, 2], 'two': [3, 4]}, 'beta': {'three': [1, 2]}, 'name': 'node1'}
        r2 = {'alpha': {'one': [1, 2], 'two': [3, 4]}, 'beta': {'three': [1, 2, 3, 4, 1, 2]}, 'name': 'node2'}
        n2 = Parameters({ 'name': 'node2' }, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.merge(p4)
        n2.interpolate()
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_nested_refs_with_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': 1, 'two': 2 } }, SETTINGS, '')
        p2 = Parameters({ 'beta': { 'three': 'one' } }, SETTINGS, '')
        p3 = Parameters({ 'beta': { 'three': 'two' } }, SETTINGS, '')
        p4 = Parameters({ 'beta': { 'four': '${alpha:${beta:three}}' } }, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1' }, SETTINGS, '')
        r1 = {'alpha': {'one': 1, 'two': 2}, 'beta': {'three': 'one', 'four': 1}, 'name': 'node1'}
        r2 = {'alpha': {'one': 1, 'two': 2}, 'beta': {'three': 'two', 'four': 2}, 'name': 'node2'}
        n1.merge(p1)
        n1.merge(p4)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({ 'name': 'node2' }, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p4)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_nested_refs_error_message(self):
        # beta is missing, oops
        p1 = Parameters({'alpha': {'one': 1, 'two': 2}, 'gamma': '${alpha:${beta}}'}, SETTINGS, '')
        with self.assertRaises(InterpolationError) as error:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Bad references, at gamma\n      ${beta}")

    def test_multiple_resolve_errors(self):
        p1 = Parameters({'alpha': '${gamma}', 'beta': '${gamma}'}, SETTINGS, '')
        with self.assertRaises(ResolveErrorList) as error:
            p1.interpolate()
        # interpolation can start with either alpha or beta
        self.assertIn(error.exception.message, [ "-> \n   Cannot resolve ${gamma}, at alpha\n   Cannot resolve ${gamma}, at beta",
                                                    "-> \n   Cannot resolve ${gamma}, at beta\n   Cannot resolve ${gamma}, at alpha"])

    def test_force_single_resolve_error(self):
        settings = copy.deepcopy(SETTINGS)
        settings.group_errors = False
        p1 = Parameters({'alpha': '${gamma}', 'beta': '${gamma}'}, settings, '')
        with self.assertRaises(ResolveError) as error:
            p1.interpolate()
        # interpolation can start with either alpha or beta
        self.assertIn(error.exception.message, [ "-> \n   Cannot resolve ${gamma}, at alpha",
                                                 "-> \n   Cannot resolve ${gamma}, at beta"])

    def test_ignore_overwriten_missing_reference(self):
        settings = copy.deepcopy(SETTINGS)
        settings.ignore_overwritten_missing_references = True
        p1 = Parameters({'alpha': '${beta}'}, settings, '')
        p2 = Parameters({'alpha': '${gamma}'}, settings, '')
        p3 = Parameters({'gamma': 3}, settings, '')
        r1 = {'alpha': 3, 'gamma': 3}
        p1.merge(p2)
        p1.merge(p3)
        err1 = "[WARNING] Reference '${beta}' undefined\n"
        with mock.patch('sys.stderr', new=MockDevice()) as std_err:
            p1.interpolate()
        self.assertEqual(p1.as_dict(), r1)
        self.assertEqual(std_err.text(), err1)

    def test_ignore_overwriten_missing_reference_last_value(self):
        # an error should be raised if the last reference to be merged
        # is missing even if ignore_overwritten_missing_references is true
        settings = copy.deepcopy(SETTINGS)
        settings.ignore_overwritten_missing_references = True
        p1 = Parameters({'alpha': '${gamma}'}, settings, '')
        p2 = Parameters({'alpha': '${beta}'}, settings, '')
        p3 = Parameters({'gamma': 3}, settings, '')
        p1.merge(p2)
        p1.merge(p3)
        with self.assertRaises(InterpolationError) as error:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Cannot resolve ${beta}, at alpha")

    def test_ignore_overwriten_missing_reference_dict(self):
        # setting ignore_overwritten_missing_references to true should
        # not change the behaviour for dicts
        settings = copy.deepcopy(SETTINGS)
        settings.ignore_overwritten_missing_references = True
        p1 = Parameters({'alpha': '${beta}'}, settings, '')
        p2 = Parameters({'alpha': '${gamma}'}, settings, '')
        p3 = Parameters({'gamma': {'one': 1, 'two': 2}}, settings, '')
        err1 = "[WARNING] Reference '${beta}' undefined\n"
        p1.merge(p2)
        p1.merge(p3)
        with self.assertRaises(InterpolationError) as error, mock.patch('sys.stderr', new=MockDevice()) as std_err:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Cannot resolve ${beta}, at alpha")
        self.assertEqual(std_err.text(), err1)

    def test_escaped_string_in_ref_dict_1(self):
        # test with escaped string in first dict to be merged
        p1 = Parameters({'a': { 'one': '${a_ref}' }, 'b': { 'two': '\${not_a_ref}' }, 'c': '${b}', 'a_ref': 123}, SETTINGS, '')
        p2 = Parameters({'c': '${a}'}, SETTINGS, '')
        r = { 'a': { 'one': 123 }, 'b': { 'two': '${not_a_ref}' }, 'c': { 'one': 123, 'two': '${not_a_ref}' }, 'a_ref': 123}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_escaped_string_in_ref_dict_2(self):
        # test with escaped string in second dict to be merged
        p1 = Parameters({'a': { 'one': '${a_ref}' }, 'b': { 'two': '\${not_a_ref}' }, 'c': '${a}', 'a_ref': 123}, SETTINGS, '')
        p2 = Parameters({'c': '${b}'}, SETTINGS, '')
        r = { 'a': { 'one': 123 }, 'b': { 'two': '${not_a_ref}' }, 'c': { 'one': 123, 'two': '${not_a_ref}' }, 'a_ref': 123}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_complex_overwrites_1(self):
        # find a better name for this test
        p1 = Parameters({ 'test': { 'dict': { 'a': '${values:one}', 'b': '${values:two}' } },
                          'values': { 'one': 1, 'two': 2, 'three': { 'x': 'X', 'y': 'Y' } } }, SETTINGS, '')
        p2 = Parameters({ 'test': { 'dict': { 'c': '${values:two}' } } }, SETTINGS, '')
        p3 = Parameters({ 'test': { 'dict': { '~b': '${values:three}' } } }, SETTINGS, '')
        r = {'test': {'dict': {'a': 1, 'b': {'x': 'X', 'y': 'Y'}, 'c': 2}}, 'values': {'one': 1, 'three': {'x': 'X', 'y': 'Y'}, 'two': 2} }
        p2.merge(p3)
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_escaped_string_overwrites(self):
        p1 = Parameters({ 'test': '\${not_a_ref}' }, SETTINGS, '')
        p2 = Parameters({ 'test': '\${also_not_a_ref}' }, SETTINGS, '')
        r = { 'test': '${also_not_a_ref}' }
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_escaped_string_in_ref_dict_overwrite(self):
        p1 = Parameters({'a': { 'one': '\${not_a_ref}' }, 'b': { 'two': '\${also_not_a_ref}' }}, SETTINGS, '')
        p2 = Parameters({'c': '${a}'}, SETTINGS, '')
        p3 = Parameters({'c': '${b}'}, SETTINGS, '')
        p4 = Parameters({'c': { 'one': '\${again_not_a_ref}' } }, SETTINGS, '')
        r = {'a': {'one': '${not_a_ref}'}, 'b': {'two': '${also_not_a_ref}'}, 'c': {'one': '${again_not_a_ref}', 'two': '${also_not_a_ref}'}}
        p1.merge(p2)
        p1.merge(p3)
        p1.merge(p4)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_strict_constant_parameter(self):
        p1 = Parameters({'one': { 'a': 1} }, SETTINGS, 'first')
        p2 = Parameters({'one': { '=a': 2} }, SETTINGS, 'second')
        p3 = Parameters({'one': { 'a': 3} }, SETTINGS, 'third')
        with self.assertRaises(ChangedConstantError) as e:
            p1.merge(p2)
            p1.merge(p3)
            p1.interpolate()
        self.assertEqual(e.exception.message, "-> \n   Attempt to change constant value, at one:a, in second; third")

    def test_constant_parameter(self):
        settings = Settings({'strict_constant_parameters': False})
        p1 = Parameters({'one': { 'a': 1} }, settings, 'first')
        p2 = Parameters({'one': { '=a': 2} }, settings, 'second')
        p3 = Parameters({'one': { 'a': 3} }, settings, 'third')
        r = {'one': { 'a': 2 } }
        p1.merge(p2)
        p1.merge(p3)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_interpolated_list_type(self):
        p1 = Parameters({'a': [ 1, 2, 3 ]}, SETTINGS, 'first')
        r = {'a': [ 1, 2, 3 ]}
        self.assertIs(type(p1.as_dict()['a']), ParameterList)
        p1.interpolate()
        self.assertIs(type(p1.as_dict()['a']), list)
        self.assertEqual(p1.as_dict(), r)

    def test_interpolated_dict_type(self):
        p1 = Parameters({'a': { 'one': 1, 'two': 2, 'three': 3 }}, SETTINGS, 'first')
        r = {'a': { 'one': 1, 'two': 2, 'three': 3 }}
        self.assertIs(type(p1.as_dict()['a']), ParameterDict)
        p1.interpolate()
        self.assertIs(type(p1.as_dict()['a']), dict)
        self.assertEqual(p1.as_dict(), r)

    def test_merged_interpolated_list_type(self):
        p1 = Parameters({'a': [ 1, 2, 3 ]}, SETTINGS, 'first')
        p2 = Parameters({'a': [ 4, 5, 6 ]}, SETTINGS, 'second')
        r = {'a': [ 1, 2, 3, 4, 5, 6 ]}
        self.assertIs(type(p1.as_dict()['a']), ParameterList)
        self.assertIs(type(p2.as_dict()['a']), ParameterList)
        p1.merge(p2)
        self.assertIs(type(p1.as_dict()['a']), ValueList)
        p1.interpolate()
        self.assertIs(type(p1.as_dict()['a']), list)
        self.assertEqual(p1.as_dict(), r)

    def test_merged_interpolated_dict_type(self):
        p1 = Parameters({'a': { 'one': 1, 'two': 2, 'three': 3 }}, SETTINGS, 'first')
        p2 = Parameters({'a': { 'four': 4, 'five': 5, 'six': 6 }}, SETTINGS, 'second')
        r = {'a': { 'one': 1, 'two': 2, 'three': 3, 'four': 4, 'five': 5, 'six': 6}}
        self.assertIs(type(p1.as_dict()['a']), ParameterDict)
        self.assertIs(type(p2.as_dict()['a']), ParameterDict)
        p1.merge(p2)
        self.assertIs(type(p1.as_dict()['a']), ParameterDict)
        p1.interpolate()
        self.assertIs(type(p1.as_dict()['a']), dict)
        self.assertEqual(p1.as_dict(), r)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

import os
import sys
import tempfile
import unittest
from shutil import copyfile

import yaml
from kapitan import defaults
from kapitan.cached import reset_cache
from kapitan.cli import main
from kapitan.errors import KubernetesManifestValidationError
from kapitan.validator.kubernetes_validator import KubernetesManifestValidator


class KubernetesValidatorTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join("examples", "kubernetes"))
        self.cache_dir = tempfile.mkdtemp()
        self.validator = KubernetesManifestValidator(self.cache_dir)

    def test_download_and_cache(self):
        kind = "service"
        version = "1.14.0"
        downloaded_schema = self.validator._get_schema_from_web(kind, version)
        self.validator._cache_schema(kind, version, downloaded_schema)
        self.assertTrue(
            os.path.isfile(os.path.join(self.cache_dir, defaults.FILE_PATH_FORMAT.format(version, kind)))
        )

    def test_load_from_cache(self):
        kind = "deployment"
        version = "1.11.0"  # different version just for testing
        downloaded_schema = self.validator._get_schema_from_web(kind, version)
        self.validator._cache_schema(kind, version, downloaded_schema)
        self.assertIsInstance(self.validator._get_cached_schema(kind, version), dict)

    def test_validate(self):
        service_manifest_string = """
            apiVersion: v1
            kind: Service
            metadata:
              name: my-service
            spec:
              selector:
                app: MyApp
              ports:
              - protocol: TCP
                port: 80
                targetPort: 9376
        """
        manifest_path = os.path.join(self.cache_dir, "service_manifest.yaml")
        with open(manifest_path, "w") as fp:
            fp.write(service_manifest_string)
        self.validator.validate([manifest_path], kind="service", version="1.14.0")

        with self.assertRaises(KubernetesManifestValidationError):
            self.validator.validate([manifest_path], kind="deployment", version="1.14.0")

    def test_validate_command_pass(self):
        sys.argv = ["kapitan", "validate", "--schemas-path", self.cache_dir]
        try:
            main()
        except SystemExit:
            self.fail("Kubernetes manifest validation error raised unexpectedly")

    def test_validate_command_fail(self):
        file_name_format = "inventory/classes/component/mysql{}.yml"
        original_file = file_name_format.format("")
        copied_file = file_name_format.format("_copy")
        copyfile(original_file, copied_file)
        wrong_manifest_kind = "deployment"
        with open(original_file, "r") as fp:
            d = yaml.safe_load(fp)
            # change kind from service to deployment
            d["parameters"]["kapitan"]["validate"][0]["kind"] = wrong_manifest_kind
        with open(original_file, "w") as fp:
            yaml.dump(d, fp, default_flow_style=False)

        sys.argv = ["kapitan", "validate", "--schemas-path", self.cache_dir]
        with self.assertRaises(SystemExit), self.assertLogs(logger="kapitan.targets", level="ERROR") as log:
            try:
                main()
            finally:
                # copy back the original file
                copyfile(copied_file, original_file)
                os.remove(copied_file)
        self.assertTrue(" ".join(log.output).find("invalid '{}' manifest".format(wrong_manifest_kind)) != -1)

    def test_validate_after_compile(self):
        sys.argv = [
            "kapitan",
            "compile",
            "-t",
            "minikube-mysql",
            "--validate",
            "--schemas-path",
            self.cache_dir,
        ]
        try:
            main()
        except SystemExit:
            self.fail("Kubernetes manifest validation error raised unexpectedly")

    def tearDown(self):
        os.chdir("../../")
        reset_cache()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

"kadet tests"

import tempfile
import unittest
from kapitan.errors import CompileError
from kapitan.inputs.kadet import BaseObj, Dict


class KadetTestObj(BaseObj):
    def new(self):
        self.need("name", "Need a name string")
        self.need("size", "Need a size int")

    def body(self):
        self.root.name = self.kwargs.name
        self.root.size = self.kwargs.size
        self.root.first_key = 1
        self.root.nested.first_key = 2
        self.root["traditional_key"] = 3
        self.root.with_dict = {"A": "dict"}
        self.root.with_baseobj_init_as = BaseObj.from_dict({"init": "as"})
        bobj = BaseObj()
        bobj.root.inside = "BaseObj"
        self.root.with_baseobj = bobj
        self.root.with_another_dict = Dict({"Another": "Dict"})


class KadetTestObjWithInner(KadetTestObj):
    def body(self):
        super().body()

        class Inner(BaseObj):  # noqa E306
            def body(self):
                self.root.i_am_inside = True

        self.root.inner = Inner()


class KadetTest(unittest.TestCase):
    def test_parse_kwargs(self):
        kobj = BaseObj.from_dict({"this": "that", "not_hidden": True})
        output = kobj.to_dict()
        desired_output = {"this": "that", "not_hidden": True}
        self.assertEqual(output, desired_output)

    def test_to_dict(self):
        kobj = KadetTestObj(name="testObj", size=5)
        output = kobj.to_dict()
        desired_output = {
            "name": "testObj",
            "size": 5,
            "first_key": 1,
            "traditional_key": 3,
            "nested": {"first_key": 2},
            "with_dict": {"A": "dict"},
            "with_baseobj_init_as": {"init": "as"},
            "with_baseobj": {"inside": "BaseObj"},
            "with_another_dict": {"Another": "Dict"},
        }
        self.assertEqual(output, desired_output)

    def test_inner(self):
        kobj = KadetTestObjWithInner(name="testWithInnerObj", size=6)
        output = kobj.to_dict()
        desired_output = {
            "name": "testWithInnerObj",
            "size": 6,
            "first_key": 1,
            "traditional_key": 3,
            "nested": {"first_key": 2},
            "with_dict": {"A": "dict"},
            "with_baseobj_init_as": {"init": "as"},
            "with_baseobj": {"inside": "BaseObj"},
            "with_another_dict": {"Another": "Dict"},
            "inner": {"i_am_inside": True},
        }
        self.assertEqual(output, desired_output)

    def test_lists(self):
        kobj = KadetTestObj(name="testObj", size=5)
        kobj.root.with_lists = [
            Dict({"i_am_inside_a_list": True}),
            BaseObj.from_dict({"me": "too"}),
            BaseObj.from_dict({"list_of_objs": [BaseObj.from_dict(dict(a=1, b=2)), Dict(dict(c=3, d=4))]}),
        ]
        output = kobj.to_dict()
        desired_output = {
            "name": "testObj",
            "size": 5,
            "first_key": 1,
            "traditional_key": 3,
            "nested": {"first_key": 2},
            "with_dict": {"A": "dict"},
            "with_baseobj_init_as": {"init": "as"},
            "with_baseobj": {"inside": "BaseObj"},
            "with_another_dict": {"Another": "Dict"},
            "with_lists": [
                {"i_am_inside_a_list": True},
                {"me": "too"},
                {"list_of_objs": [{"a": 1, "b": 2}, {"c": 3, "d": 4}]},
            ],
        }
        self.assertEqual(output, desired_output)

    def test_need(self):
        with self.assertRaises(CompileError):
            KadetTestObj(this_should_error=True)

    def test_update_root_yaml(self):
        yaml_file = tempfile.mktemp(suffix=".yml")
        with open(yaml_file, "w") as fp:
            fp.write("this: that\nlist: [1,2,3]\n")

        class KadetObjFromYaml(BaseObj):
            def new(self):
                self.update_root(yaml_file)

        output = KadetObjFromYaml().to_dict()
        desired_output = {"this": "that", "list": [1, 2, 3]}
        self.assertEqual(output, desired_output)

    def test_update_root_json(self):
        json_file = tempfile.mktemp(suffix=".json")
        with open(json_file, "w") as fp:
            fp.write('{"this": "that", "list": [1,2,3]}')

        class KadetObjFromYaml(BaseObj):
            def new(self):
                self.update_root(json_file)

        output = KadetObjFromYaml().to_dict()
        desired_output = {"this": "that", "list": [1, 2, 3]}
        self.assertEqual(output, desired_output)

    def test_from_json(self):
        json_file = tempfile.mktemp()
        with open(json_file, "w") as fp:
            fp.write('{"this": "that", "list": [1,2,3]}')

        kobj = BaseObj.from_json(json_file)
        output = kobj.to_dict()
        desired_output = {"this": "that", "list": [1, 2, 3]}
        self.assertEqual(output, desired_output)

    def test_from_yaml(self):
        yaml_file = tempfile.mktemp()
        with open(yaml_file, "w") as fp:
            fp.write("this: that\nlist: [1,2,3]\n")

        kobj = BaseObj.from_yaml(yaml_file)
        output = kobj.to_dict()
        desired_output = {"this": "that", "list": [1, 2, 3]}
        self.assertEqual(output, desired_output)
